/*
 * CvsGraph graphical representation generator of brances and revisions
 * of a file in cvs/rcs.
 *
 * Copyright (C) 2001  B. Stultiens
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Approx. layout of a cvs/rcs log:
 *
 * ws		::= "[ \t]*"
 * rev_nr	::= "[:digit:]+(\.[:digit:]+)*"
 * path_name	::= "/?(([^\n/]*)/)+"
 * file_name	::= "[^\n]+"
 * file_path	::= "{file_path}{file_name}"
 * tag		::= "[^,.$@:;\0-\037]+"
 * number	::= "[:digit:]+"
 * separator	::= "(-{28})|(={78)\n"
 *
 * The header is identified with this snippet until
 * a {separator} is encountered:
 *	"RCS file:{ws}{file_path}"
 *	"Working file:{ws}{file_name}"
 *	"head:{ws}{rev_nr}"
 *	"branch:{ws}{rev_nr}?"
 *	"locks:{ws}[^\n]*"
 *	"access list:{ws}[^\n]*"
 *	"symbolic names:"
 *	"(\t{tag}:{rev_nr}\n)*"
 *	"keyword substitution:{ws}[^\n]*"
 *	"total revisions:{ws}{number};{ws}selected revisions:{ws}{number}"
 *	"description:"
 *	"<any text you can imagine until a separator>"
 *
 * Each revision is identiefied with:
 *	"{separator}"
 *	"revision{ws}{rev_nr}"
 *	"date: 2001/02/15 20:17:37;  author: bertho;  state: Exp;  lines: +2 -0
 *	"any text as a comment until a separator>"
 *
 * The last revision has the "={78}" separator. Eventually, a next file may be
 * appended.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <regex.h>
#include <errno.h>
#include <getopt.h>
 
#include <gd.h>
#include <gdfontt.h>

#include "config.h"
#include "cvsgraph.h"
#include "utils.h"
#include "readconf.h"

#if !defined(HAVE_IMAGE_GIF) && !defined(HAVE_IMAGE_PNG) && !defined(HAVE_IMAGE_JPEG)
# error No image output format available. Check libgd
#endif


/*#define DEBUG		1*/

#define RLOGCMD		"/usr/bin/rlog"
#define DEVNULL		"/dev/null"
#define CONFFILENAME	"cvsgraph.conf"

#ifndef ETCDIR
# define ETCDIR		"/usr/local/etc"
#endif

#ifndef MAX
# define MAX(a,b)	((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
# define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

#define ALIGN_HL	0x00
#define ALIGN_HC	0x01
#define ALIGN_HR	0x02
#define ALIGN_HX	0x0f 
#define ALIGN_VT	0x00
#define ALIGN_VC	0x10
#define ALIGN_VB	0x20
#define ALIGN_VX	0xf0
 
typedef struct __revid_t
{
	char	*branch;
	char	*rev;
	int	isbranch;
} revid_t;

typedef struct __tag_t
{
	char	*tag;
	revid_t	*rev;
} tag_t;

struct __branch_t;

typedef struct __revision_t
{
	revid_t		*rev;
	char		*info;
	char		*comment;
	tag_t		**tags;
	int		ntags;
	struct __branch_t	**branches;
	int		nbranches;
	int		w, h;
	int		x, y;
} revision_t;

typedef struct __branch_t
{
	char		*branch;
	revision_t	*branchpoint;
	tag_t		*tag;
	revision_t	**revs;
	int		nrevs;
	int		tw, th;
	int		w, h;
	int		x, y;
} branch_t;

typedef struct __rcsfilelog_t
{
	char		*path;
	char		*name;
	revid_t		*head;
	char		*branch;
	char		*locks;
	char		*access;
	char		*keyword;
	char		*totalrevs;
	char		*comment;
	tag_t		**tags;
	int		ntags;
	revision_t	**revs;
	int		nrevs;
	branch_t	**branches;
	int		nbranches;
	int		tw, th;
} rcsfilelog_t;

/*
 **************************************************************************
 * Globals
 **************************************************************************
 */

char *rlogcmd = RLOGCMD;
char *devnull = DEVNULL;

config_t conf;

/*
 **************************************************************************
 * Debug routines
 **************************************************************************
 */
#ifdef DEBUG
void dump_revid(const char *s, revid_t *r)
{
	fprintf(stderr, "%s.branch  : '%s'\n", s, r->branch);
	fprintf(stderr, "%s.rev     : '%s'\n", s, r->rev);
	fprintf(stderr, "%s.isbranch: %d\n", s, r->isbranch);
}

void dump_tag(const char *s, tag_t *t)
{
	fprintf(stderr, "%s", s);
	dump_revid(t->tag, t->rev);
}

void dump_rev(revision_t *r)
{
	int i;
	dump_revid("Revision", r->rev);
	fprintf(stderr, "Revision.Info   : '%s'\n", r->info);
	fprintf(stderr, "Revision.Comment: '%s'\n", r->comment);
	for(i = 0; i < r->ntags; i++)
		dump_tag("Revision.Tag: ", r->tags[i]);
}

void dump_branch(branch_t *b)
{
	int i;
	fprintf(stderr, "Branch: '%s'\n", b->branch);
	if(b->tag)
		dump_tag("branchtag:", b->tag);
	for(i = 0; i < b->nrevs; i++)
		fprintf(stderr, "Branch.Rev: '%s'\n", b->revs[i]->rev->rev);
}

void dump_log(rcsfilelog_t *r)
{
	int i;

	fprintf(stderr, "Path   : '%s'\n", r->path);
	fprintf(stderr, "Name   : '%s'\n", r->name);
	dump_revid("Head", r->head);
	fprintf(stderr, "Branch : '%s'\n", r->branch);
	fprintf(stderr, "Locks  : '%s'\n", r->locks);
	fprintf(stderr, "Access : '%s'\n", r->access);
	fprintf(stderr, "Keyword: '%s'\n", r->keyword);
	fprintf(stderr, "Total  : '%s'\n", r->totalrevs);
	fprintf(stderr, "Comment: '%s'\n", r->comment);
	for(i = 0; i < r->ntags; i++)
		dump_tag("", r->tags[i]);
	for(i = 0; i < r->nrevs; i++)
		dump_rev(r->revs[i]);
	for(i = 0; i < r->nbranches; i++)
		dump_branch(r->branches[i]);
}
#endif

/*
 **************************************************************************
 * Retrieve the log entries
 **************************************************************************
 */
FILE *get_log(const char *cvsroot, const char *module, const char *file)
{
	pid_t pid;
	int nul;
	FILE *tmp;
	char *cmd = NULL;
	int status;
	mode_t um;

	if((nul = open(devnull, O_RDWR, S_IRUSR|S_IWUSR)) == -1)
		return NULL;

	um = umask(0177);	/* Set tempfiles to max 0600 permissions */
	if((tmp = tmpfile()) == NULL)
	{
		close(nul);
		return NULL;
	}
	umask(um);

	cmd = xmalloc(strlen(cvsroot) + + strlen(module) + strlen(file) + 2 + 1);
	sprintf(cmd, "%s/%s/%s", cvsroot, module, file);

	switch(pid = fork())
	{
	case -1:	/* Error */
		close(nul);
		fclose(tmp);
		xfree(cmd);
		return NULL;
	case 0:		/* Child */
		if((dup2(nul, STDIN_FILENO)) == -1)	exit(126);
		if((dup2(fileno(tmp), STDOUT_FILENO)) == -1)	exit(126);
		if((dup2(nul, STDERR_FILENO)) == -1)	exit(126);
		close(nul);
		fclose(tmp);
		execl(rlogcmd, rlogcmd, cmd, NULL);
		exit(127);
		break;
	default:	/* Parent */
		close(nul);
		xfree(cmd);
		while(1)
		{
			if(waitpid(pid, &status, 0) == -1)
			{
				if(errno != EINTR)
				{
					fclose(tmp);
					return NULL;
				}
			}
			else
				break;
		}
		break;
	}

	if(WIFEXITED(status) && WEXITSTATUS(status) == 0)
	{
		if(fseek(tmp, 0, SEEK_SET) != (off_t)-1)
		{
			return tmp;
		}
		else
		{
			fclose(tmp);
			return NULL;
		}
	}
	else
		fclose(tmp);
	return NULL;
}

/*
 **************************************************************************
 * Parse the log entries
 **************************************************************************
 */
char *strip_dup(const char *s)
{
	int l = strlen(s);
	char *c = xmalloc(l+1);

	strcpy(c, s);
	while(*c == ' ' || *c == '\t')
	{
		memmove(c, c+1, l--);
	}
	while(l && strchr(" \t\r\n", c[l]))
		c[l--] = '\0';
	return c;
}

revid_t *make_revid(const char *s)
{
	char *c = strip_dup(s);
	char *cptr;
	int dots = 0;
	revid_t *r = xmalloc(sizeof(*r));
	for(cptr = c; *cptr; cptr++)
	{
		if(*cptr == '.')
			dots++;
	}
	if(!dots)
	{
		r->rev = xstrdup("");
		r->branch = xstrdup(s);
		r->isbranch = 1;
	}
	else if(!*c)
	{
		r->rev = xstrdup("?.?");
		r->branch = xstrdup("?");
	}
	else if(dots & 1)
	{
		char *t;
		r->rev = c;
		r->branch = xstrdup(c);
		cptr = strrchr(r->branch, '.');
		assert(cptr != NULL);
		*cptr = '\0';
		t = strrchr(r->branch, '.');
		if((t&& !strcmp(t+1, "0")) || (!t && !strcmp(r->branch, "0")))
		{
			/* Magic branch numbers "x.x.0.x" */
			r->isbranch = 1;
			if(t)
				strcpy(t+1, cptr+1);
			else
				strcpy(r->branch, cptr+1);
		}
	}
	else
	{
		r->isbranch = 1;
		r->branch = c;
		r->rev = xmalloc(strlen(c) + 3);
		strcpy(r->rev, c);
		strcat(r->rev, ".?");
	}
	return r;
}

char *add_comment(char *c, const char *n)
{
	int l;
	char *r;
	assert(n != NULL);
	l = strlen(n);
	if(!c)
	{
		r = xmalloc(l+1);
		strcpy(r, n);
	}
	else
	{
		r = xmalloc(l+strlen(c)+1+1);
		strcpy(r, c);
		strcat(r, "\n");
		strcat(r, n);
	}
	return r;
}

int get_line(FILE *fp, char *buf, int maxlen)
{
	int n;
	int seennl;
retry:
	seennl = 0;
	if(!fgets(buf, maxlen, fp))
		return feof(fp) ? 0 : -1;
	n = strlen(buf);
	while(n && buf[n-1] == '\n')
	{
		seennl = 1;
		buf[--n] = '\0';
	}
	if(!n)
		goto retry;
	if(!seennl)
	{
		while(fgetc(fp) != '\n')
			;
	}
	return n;
}

rcsfilelog_t *parse_log(FILE *fp)
{
	rcsfilelog_t *p;
	int state = 0;
	regex_t rerev;
	regex_t reinfo;

	if(regcomp(&rerev, "^revision[ \\t]*[0-9]+(\\.[0-9]+)*", REG_EXTENDED))
		return NULL;
	if(regcomp(&reinfo, "^date:[^;]*;[ \\t]*author:[^;]*;[ \\t]+state:", REG_EXTENDED))
	{
		regfree(&rerev);
		return NULL;
	}
	p = xmalloc(sizeof(*p));
	while(state != 4)
	{
		char buf[256];
		int n;
		n = get_line(fp, buf, sizeof(buf));
		if(!n)
			break;
		if(n == -1)
		{
			perror("tempfile read");
			xfree(p);
			regfree(&rerev);
			regfree(&reinfo);
			return NULL;
		}
		switch(state)
		{
		case 0:	/* Prologue */
more_prologue:
			if(!strncmp(buf, "RCS file:", 9))
			{
				p->path = strip_dup(buf+9);
			}
			else if(!strncmp(buf, "Working file:", 13))
			{
				p->name = strip_dup(buf+13);
			}
			else if(!strncmp(buf, "head:", 5))
			{
				p->head = make_revid(buf+5);
			}
			else if(!strncmp(buf, "branch:", 7))
			{
				p->branch = strip_dup(buf+7);
			}
			else if(!strncmp(buf, "locks:", 6))
			{
				p->locks = strip_dup(buf+6);
			}
			else if(!strncmp(buf, "access list:", 12))
			{
				p->access = strip_dup(buf+12);
			}
			else if(!strncmp(buf, "keyword substitution:", 21))
			{
				p->keyword = strip_dup(buf+21);
			}
			else if(!strncmp(buf, "total revisions:", 16))
			{
				p->totalrevs = strip_dup(buf+16);
			}
			else if(!strncmp(buf, "description:", 12))
			{
				state = 2;
			}
			else if(!strncmp(buf, "symbolic names:", 15))
			{
				state = 1;
			}
			else
			{
				fprintf(stderr, "Unknown keyword(s) in line '%s' (state=%d)\n", buf, state);
				xfree(p);
				return NULL;
			}
			break;
		case 1:	/* Tags */
			if(*buf != '\t')
			{
				state = 0;
				goto more_prologue;
			}
			else
			{
				char *rev = strrchr(buf, ':');
				tag_t *t;
				if(!rev)
				{
					state = 2;
					goto more_prologue;
				}
				*rev = '\0';
				t = xmalloc(sizeof(*t));
				t->tag = strip_dup(buf);
				t->rev = make_revid(rev+1);
				p->tags = xrealloc(p->tags, (p->ntags+1) * sizeof(p->tags[0]));
				p->tags[p->ntags] = t;
				p->ntags++;
			}
			break;
		case 2:	/* Description */
add_description:
			if(!strcmp(buf, "----------------------------"))
			{
				/* End of description */
				state = 3;
				break;
			}
			else if(!strcmp(buf, "============================================================================="))
			{
				/* End of log */
				state = 4;
				break;
			}
			if(!p->nrevs)
				p->comment = add_comment(p->comment, buf);
			else
				p->revs[p->nrevs-1]->comment = add_comment(p->revs[p->nrevs-1]->comment, buf);
			break;
		case 3:
			if(!regexec(&rerev, buf, 0, NULL, 0))
			{
				revision_t *r = xmalloc(sizeof(*r));
				p->revs = xrealloc(p->revs, (p->nrevs+1) * sizeof(p->revs[0]));
				p->revs[p->nrevs] = r;
				p->nrevs++;
				r->rev = make_revid(buf+8);
			}
			else if(!regexec(&reinfo, buf, 0, NULL, 0))
			{
				assert(p->nrevs > 0);
				p->revs[p->nrevs-1]->info = strip_dup(buf);
			}
			else
			{
				/* No match means the description/comment */
				state = 2;
				goto add_description;
			}
			break;
		default:
			fprintf(stderr, "Illegal state (%d) in parser\n", state);
			xfree(p);
			regfree(&rerev);
			regfree(&reinfo);
			return NULL;
		}
	}
	regfree(&rerev);
	regfree(&reinfo);
	return p;
}

/*
 **************************************************************************
 * Sort and find helpers
 **************************************************************************
 */
int count_dots(const char *s)
{
	int i;
	for(i = 0; *s; s++)
	{
		if(*s == '.')
			i++;
	}
	return i;
}

int compare_revid(const revid_t *r1, const revid_t *r2)
{
	int d1, d2;
	char *v1, *v2;
	char *s1, *s2;
	int retval = 0;
	assert(r1 != NULL);
	assert(r1->rev != NULL);
	assert(r2 != NULL);
	assert(r2->rev != NULL);

	d1 = count_dots(r1->rev);
	d2 = count_dots(r2->rev);
	if(d1 != d2)
	{
		return d1 - d2;
	}

	s1 = v1 = xstrdup(r1->rev);
	s2 = v2 = xstrdup(r2->rev);
	while(1)
	{
		char *vc1 = strchr(s1, '.');
		char *vc2 = strchr(s2, '.');
		if(vc1 && vc2)
			*vc1 = *vc2 = '\0';
		if(*s1 && *s2)
		{
			d1 = atoi(s1);
			d2 = atoi(s2);
			if(d1 != d2)
			{
				retval = d1 - d2;
				break;
			}
		}
		if(!vc1 || !vc2)
			break;
		s1 = vc1 + 1;
		s2 = vc2 + 1;
	}
	xfree(v1);
	xfree(v2);
	return retval;
}

int tag_sort(const void *t1, const void *t2)
{
#define TAGPTR(t)	(*((tag_t **)t))
	return strcmp(TAGPTR(t1)->rev->rev, TAGPTR(t2)->rev->rev);
#undef TAGPTR
}

int rev_sort(const void *v1, const void *v2)
{
#define REVPTR(t)	(*((revision_t **)t))
	return compare_revid(REVPTR(v1)->rev, REVPTR(v2)->rev);
#undef REVPTR
}

int branch_sort(const void *b1, const void *b2)
{
#define BPTR(t)	(*((branch_t **)t))
	return strcmp(BPTR(b1)->branch, BPTR(b2)->branch);
#undef BPTR
}

int rev_cmp(const void *id, const void *v)
{
#define REVPTR(t)	(*((revision_t **)t))
	return compare_revid((revid_t *)id, REVPTR(v)->rev);
#undef REVPTR
}

revision_t *find_revision(rcsfilelog_t * rcs, revid_t *id)
{
	revision_t **r;
	if(id->isbranch)
		return NULL;
	r = bsearch(id, rcs->revs, rcs->nrevs, sizeof(rcs->revs[0]), rev_cmp);
	if(!r)
		return NULL;
	else
		return *r;
}

int branch_cmp(const void *s, const void *b)
{
	return strcmp((const char *)s, (*((branch_t **)b))->branch);
}

branch_t *find_branch(rcsfilelog_t *rcs, const char *id)
{
	branch_t **b;
	b = bsearch(id, rcs->branches, rcs->nbranches, sizeof(rcs->branches[0]), branch_cmp);
	if(!b)
		return NULL;
	else
		return *b;
}

tag_t *find_branchtag(rcsfilelog_t * rcs, const char *id)
{
	int i;
	for(i = 0; i < rcs->ntags; i++)
	{
		if(!rcs->tags[i]->rev->isbranch)
			continue;
		if(!strcmp(id, rcs->tags[i]->rev->branch))
			return rcs->tags[i];
	}
	return NULL;
}

/*
 **************************************************************************
 * Drawing routines
 **************************************************************************
 */
int get_swidth(const char *s, font_t *f)
{
	if(!s)
		return 0;
	return strlen(s) * (*f)->w;
}

int get_sheight(const char *s, font_t *f)
{
	int nl;
	if(!s)
		return 0;
	for(nl = 1; *s; s++)
	{
		if(*s == '\n' && s[1])
			nl++;
	}
	return nl * (*f)->h;
}

void draw_rbox(gdImagePtr im, int x1, int y1, int x2, int y2, int r, color_t *color)
{
	int r2 = 2*r;
	gdImageLine(im, x1+r, y1, x2-r, y1, color->id);
	gdImageLine(im, x1+r, y2, x2-r, y2, color->id);
	gdImageLine(im, x1, y1+r, x1, y2-r, color->id);
	gdImageLine(im, x2, y1+r, x2, y2-r, color->id);
	if(r)
	{
		gdImageArc(im, x1+r, y1+r, r2, r2, 180, 270, color->id);
		gdImageArc(im, x2-r, y1+r, r2, r2, 270, 360, color->id);
		gdImageArc(im, x1+r, y2-r, r2, r2,  90, 180, color->id);
		gdImageArc(im, x2-r, y2-r, r2, r2,   0,  90, color->id);
	}
}

void draw_string(gdImagePtr im, char *s, font_t *f, int x, int y, int align, color_t *c)
{
	int xx, yy;
	switch(align & ALIGN_HX)
	{
	default:
	case ALIGN_HL: xx = 0; break;
	case ALIGN_HC: xx = -get_swidth(s, f)/2; break;
	case ALIGN_HR: xx = -get_swidth(s, f); break;
	}
	switch(align & ALIGN_VX)
	{
	default:
	case ALIGN_VT: yy = 0; break;
	case ALIGN_VC: yy = -get_sheight(s, f)/2; break;
	case ALIGN_VB: yy = -get_sheight(s, f); break;
	}
	gdImageString(im, *f, x+xx+1, y+yy, s, c->id);
}

void draw_rev(gdImagePtr im, int cx, int ty, revision_t *r)
{
	int lx = cx - r->w/2;
	int rx = lx + r->w;
	int i;
	draw_rbox(im, lx, ty, rx, ty+r->h, 0, &conf.rev_color);
	ty += conf.rev_tspace;
	draw_string(im, r->rev->rev, &conf.rev_font, cx, ty, ALIGN_HC, &conf.rev_color);
	ty += get_sheight(r->rev->rev, &conf.rev_font);
	for(i = 0; i < r->ntags; i++)
	{
		draw_string(im, r->tags[i]->tag, &conf.tag_font, cx, ty, ALIGN_HC, &conf.tag_color);
		ty += get_sheight(r->tags[i]->tag, &conf.tag_font);
	}
}

void draw_branch(gdImagePtr im, int cx, int ty, branch_t *b)
{
	int lx = cx - b->w/2;
	int rx = lx + b->w;
	int yy;
	int i;
	draw_rbox(im, lx, ty, rx, ty+b->h, 5, &conf.branch_color);
	yy = conf.branch_tspace;
	draw_string(im, b->branch, &conf.branch_font, cx, ty+yy, ALIGN_HC, &conf.branch_color);
	yy += get_sheight(b->branch, &conf.branch_font);
	if(b->tag)
	{
		draw_string(im, b->tag->tag, &conf.branch_font, cx, ty+yy, ALIGN_HC, &conf.branch_color);
	}

	ty += b->h;
	for(i = 0; i < b->nrevs; i++)
	{
		gdImageLine(im, cx, ty, cx, ty+conf.rev_minline, conf.rev_color.id);
		ty += conf.rev_minline;
		draw_rev(im, cx, ty, b->revs[i]);
		ty += b->revs[i]->h;
	}
}

static char *_title;
static int _ntitle;
static int _natitle;

void add_title_str(const char *s)
{
	int l = strlen(s) + 1;
	if(_ntitle + l > _natitle)
	{
		_natitle += 128;
		_title = xrealloc(_title, _natitle * sizeof(_title[0]));
	}
	memcpy(_title+_ntitle, s, l);
	_ntitle += l-1;
}

void add_title_ch(int ch)
{
	char buf[2];
	buf[0] = ch;
	buf[1] = '\0';
	add_title_str(buf);
}

char *expand_title(rcsfilelog_t *rcs)
{
	char nb[32];
	char nr[32];
	char *cptr;

	sprintf(nb, "%d", rcs->nbranches);
	sprintf(nr, "%d", rcs->nrevs);
	for(cptr = conf.title; *cptr; cptr++)
	{
		if(*cptr == '%')
		{
			switch(*++cptr)
			{
			case 'c': add_title_str(conf.cvsroot); break;
			case 'f': add_title_str(rcs->name); break;
			case 'm': add_title_str(conf.cvsmodule); break;
			case 'r': add_title_str(nr); break;
			case 'b': add_title_str(nb); break;
			case '%': add_title_ch('%'); break;
			default:
				add_title_ch('%');
				add_title_ch(*cptr);
				break;
			}
		}
		else
			add_title_ch(*cptr);
	}
	return _title;
}

void draw_title(gdImagePtr im, char *title)
{
	char *t;
	char *s = title;
	int x = conf.title_x;
	int y = conf.title_y;
	do
	{
		t = strchr(s, '\n');
		if(t)
			*t = '\0';
		draw_string(im, s, &conf.title_font, x, y, conf.title_align, &conf.title_color);
		y += get_sheight(s, &conf.title_font);
		s = t+1;
	} while(t);
}

void draw_connector(gdImagePtr im, branch_t *b)
{
	revision_t *r = b->branchpoint;
	int x1 = r->x + r->w/2 + 2;
	int y1 = r->y + r->h/2;
	int x2 = b->x;
	int y2 = b->y;
	gdImageLine(im, x1, y1, x2, y1, conf.branch_color.id);
	gdImageLine(im, x2, y1, x2, y2, conf.branch_color.id);
}

gdImagePtr make_image(rcsfilelog_t *rcs)
{
	gdImagePtr im;
	int i;

	im = gdImageCreate(rcs->tw+conf.margin_left+conf.margin_right, rcs->th+conf.margin_top+conf.margin_bottom);
	conf.color_bg.id = gdImageColorAllocate(im, conf.color_bg.r, conf.color_bg.g, conf.color_bg.b);
	conf.tag_color.id = gdImageColorAllocate(im, conf.tag_color.r, conf.tag_color.g, conf.tag_color.b);
	conf.rev_color.id = gdImageColorAllocate(im, conf.rev_color.r, conf.rev_color.g, conf.rev_color.b);
	conf.branch_color.id = gdImageColorAllocate(im, conf.branch_color.r, conf.branch_color.g, conf.branch_color.b);
	conf.branch_bgcolor.id = gdImageColorAllocate(im, conf.branch_bgcolor.r, conf.branch_bgcolor.g, conf.branch_bgcolor.b);
	conf.title_color.id = gdImageColorAllocate(im, conf.title_color.r, conf.title_color.g, conf.title_color.b);

	for(i = 0; i < rcs->nbranches; i++)
		draw_branch(im, rcs->branches[i]->x, rcs->branches[i]->y, rcs->branches[i]);
	for(i = 0; i < rcs->nbranches; i++)
	{
		if(rcs->branches[i]->branchpoint)
			draw_connector(im, rcs->branches[i]);
	}
	draw_title(im, expand_title(rcs));

	return im;
}

void move_branch(branch_t *b, int x, int y)
{
	int i;
	b->x += x;
	b->y += y;
	for(i = 0; i < b->nrevs; i++)
	{
		b->revs[i]->x += x;
		b->revs[i]->y += y;
	}
}

void rect_union(int *x, int *y, int *w, int *h, branch_t *b)
{
	int x1 = *x;
	int x2 = x1 + *w;
	int y1 = *y;
	int y2 = y1 + *h;
	int xx1 = b->x - b->tw/2;
	int xx2 = xx1 + b->tw;
	int yy1 = b->y;
	int yy2 = yy1 + b->th;
	x1 = MIN(x1, xx1);
	x2 = MAX(x2, xx2);
	y1 = MIN(y1, yy1);
	y2 = MAX(y2, yy2);
	*x = x1;
	*y = y1;
	*w = x2 - x1;
	*h = y2 - y1;
}

void make_layout(rcsfilelog_t *rcs)
{
	int i, j;
	int x, y;
	int w, h;
	int w2;

	/* Calculate the box-sizes of the revisions */
	for(i = 0; i < rcs->nrevs; i++)
	{
		revision_t *rp;
		int w;
		int h;
		rp = rcs->revs[i];
		w = get_swidth(rp->rev->rev, &conf.rev_font);
		h = get_sheight(rp->rev->rev, &conf.rev_font);
		for(j = 0; j < rp->ntags; j++)
		{
			int ww = get_swidth(rp->tags[j]->tag, &conf.tag_font);
			if(ww > w) w = ww;
			h += get_sheight(rp->tags[j]->tag, &conf.tag_font) + conf.rev_separator;
		}
		rp->w = w + conf.rev_lspace + conf.rev_rspace;
		rp->h = h + conf.rev_tspace + conf.rev_bspace;
	}

	/* Calculate the box-sizes of the branches */
	for(i = 0; i < rcs->nbranches; i++)
	{
		branch_t *bp = rcs->branches[i];
		int w;
		int h;
		w = get_swidth(bp->branch, &conf.branch_font);
		h = get_sheight(bp->branch, &conf.branch_font);
		if(bp->tag)
		{
			int ww = get_swidth(bp->tag->tag, &conf.branch_font);
			if(ww > w) w = ww;
			h += get_sheight(bp->tag->tag, &conf.branch_font) + conf.branch_separator;
		}
		w += conf.branch_lspace + conf.branch_rspace;
		h += conf.branch_tspace + conf.branch_bspace;
		bp->w = w;
		bp->h = h;
		for(j = 0; j < bp->nrevs; j++)
		{
			if(bp->revs[j]->w > w)
				w = bp->revs[j]->w;
			h += bp->revs[j]->h + conf.rev_minline;
		}
		bp->th = h;
		bp->tw = w;
	}

	/* Calculate the relative positions of revs in a branch */
	for(i = 0; i < rcs->nbranches; i++)
	{
		branch_t *b = rcs->branches[i];
		x = b->tw/2;
		y = b->h;
		b->x = x;
		b->y = 0;
		for(j = 0; j < b->nrevs; j++)
		{
			y += conf.rev_minline;
			b->revs[j]->x = x;
			b->revs[j]->y = y;
			y += b->revs[j]->h;
		}
	}

	/* Reposition the branches FIXME: Should be recursive on branchpoint revisions within branches... */
	x = rcs->branches[0]->x;
	w2 = rcs->branches[0]->tw / 2;
	for(i = rcs->branches[0]->nrevs-1; i >= 0; i--)
	{
		revision_t *r = rcs->branches[0]->revs[i];
		for(j = 0; j < r->nbranches; j++)
		{
			branch_t *b = r->branches[j];
			x += w2 + conf.rev_minline + b->tw/2 - b->x;
			w2 = b->tw/2;
			move_branch(b, x, r->y + r->h);
			x = b->x;
		}
	}

	for(i = 0; i < rcs->nbranches; i++)
		move_branch(rcs->branches[i], conf.margin_left, conf.margin_top);

	/* Calculate overall image size */
	x = rcs->branches[0]->x - rcs->branches[0]->tw/2;
	y = rcs->branches[0]->y;
	w = rcs->branches[0]->tw;
	h = rcs->branches[0]->th;
	for(i = 1; i < rcs->nbranches; i++)
		rect_union(&x, &y, &w, &h, rcs->branches[i]);
	rcs->tw = w;
	rcs->th = h;
}

/*
 **************************************************************************
 * Configuration
 **************************************************************************
 */
int read_config(const char *path)
{
	FILE *fp;
	int r;
	if(path)
	{
		if((fp = fopen(path, "r")) == NULL)
		{
			return 0;
		}
	}
	else
	{
		if((fp = fopen("./" CONFFILENAME, "r")) == NULL)
		{
			if((fp = fopen(ETCDIR "/" CONFFILENAME, "r")) == NULL)
			{
				return 0;
			}
		}
	}

	yyin = fp;
	r = yyparse();
	fclose(fp);
	return r == 0;
}

/*
 **************************************************************************
 * Program entry
 **************************************************************************
 */
static const char usage_str[] =
	"Usage: cvsgraph [options] <file>\n"
	"  -c <file>  Read alternative config from <file>\n"
	"  -h         This message\n"
	"  -m <mod>   Use <mod> as cvs module\n"
	"  -o <file>  Output to <file>\n"
	"  -r <path>  Use <path> as cvsroot path\n"
	"  -V         Print version and exit\n"
	;

#define VERSION_STR	"1.0.0"
#define NOTICE_STR	"Copyright (c) 2001 B.Stultiens"

void add_tag(rcsfilelog_t *rcs, const char *tag, const char *rev)
{
	rcs->tags = xrealloc(rcs->tags, (rcs->ntags+1)*sizeof(rcs->tags[0]));
	rcs->tags[rcs->ntags] = xmalloc(sizeof(tag_t));
	rcs->tags[rcs->ntags]->tag = strip_dup(tag);
	rcs->tags[rcs->ntags]->rev = make_revid(rev);
	rcs->ntags++;
}

int main(int argc, char *argv[])
{
	int optc;
	char *confpath = NULL;
	char *outfile = NULL;
	char *cvsroot = NULL;
	char *cvsmodule = NULL;
	int lose = 0;
	FILE *fp;
	int n;
	rcsfilelog_t *rcs;
	gdImagePtr im;

#ifdef DEBUG
	setvbuf(stdout, NULL, 0, _IONBF);
	setvbuf(stderr, NULL, 0, _IONBF);
#endif

	while((optc = getopt(argc, argv, "c:hm:o:r:V")) != EOF)
	{
		switch(optc)
		{
		case 'c':
			confpath = xstrdup(optarg);
			break;
		case 'm':
			cvsmodule = xstrdup(optarg);
			break;
		case 'o':
			outfile = xstrdup(optarg);
			break;
		case 'r':
			cvsroot = xstrdup(optarg);
			break;
		case 'V':
			fprintf(stdout, "cvsgraph v%s, %s\n", VERSION_STR, NOTICE_STR);
			return 0;
		case 'h':
			fprintf(stdout, "%s", usage_str);
			return 0;
		default:
			lose++;
		}
	}

	if(optind >= argc)
	{
		fprintf(stderr, "Missing inputfile\n");
		lose++;
	}

	if(lose)
	{
		fprintf(stderr, "%s", usage_str);
		return 1;
	}

	/* Set defaults */
	if(!conf.tag_font)	conf.tag_font = gdFontTiny;
	if(!conf.rev_font)	conf.rev_font = gdFontTiny;
	if(!conf.branch_font)	conf.branch_font = gdFontTiny;
	if(!conf.title_font)	conf.title_font = gdFontTiny;

	if(!read_config(confpath))
	{
		fprintf(stderr, "Error reading config file\n");
		return 1;
	}

	/* Set overrides */
	if(cvsroot)		conf.cvsroot = cvsroot;
	if(cvsmodule)		conf.cvsmodule = cvsmodule;

	if((fp = get_log(conf.cvsroot, conf.cvsmodule, argv[optind])) == NULL)
	{
		fprintf(stderr, "Error getting log for '%s'\n", argv[optind]);
		return 1;
	}

	rcs = parse_log(fp);
	if(!rcs)
	{
		fprintf(stderr, "Error parsing log\n");
		return 1;
	}
	fclose(fp);

	/* Add implicit tags */
	add_tag(rcs, "HEAD", rcs->head->rev);
	add_tag(rcs, "MAIN", "1");

	/* We now have the log. Sort and reorganize a little */
	qsort(rcs->tags, rcs->ntags, sizeof(rcs->tags[0]), tag_sort);
	qsort(rcs->revs, rcs->nrevs, sizeof(rcs->revs[0]), rev_sort);

	/* Assign tags to revisions */
	for(n = 0; n < rcs->ntags; n++)
	{
		revision_t *r = find_revision(rcs, rcs->tags[n]->rev);
		if(!r)
			continue;
		r->tags = xrealloc(r->tags, (r->ntags+1) * sizeof(r->tags[0]));
		r->tags[r->ntags] = rcs->tags[n];
		r->ntags++;
	}

	/* Isolate the branches */
	for(n = 0; n < rcs->nrevs; n++)
	{
		branch_t *b = find_branch(rcs, rcs->revs[n]->rev->branch);
		if(!b)
		{
			rcs->branches = xrealloc(rcs->branches, (rcs->nbranches+1) * sizeof(rcs->branches[0]));
			b = xmalloc(sizeof(*b));
			b->branch = xstrdup(rcs->revs[n]->rev->branch);
			b->tag = find_branchtag(rcs, rcs->revs[n]->rev->branch);
			rcs->branches[rcs->nbranches] = b;
			rcs->nbranches++;
			qsort(rcs->branches, rcs->nbranches, sizeof(rcs->branches[0]), branch_sort);
		}
		b->revs = xrealloc(b->revs, (b->nrevs+1) * sizeof(b->revs[0]));
		b->revs[b->nrevs] = rcs->revs[n];
		b->nrevs++;
	}

	/* Find the branchpoints */
	for(n = 0; n < rcs->nbranches; n++)
	{
		char *prev = xstrdup(rcs->branches[n]->branch);
		char *cptr = strrchr(prev, '.');
		revision_t *r;
		revid_t rid;
		if(!cptr)	/* Main branch number */
			continue;
		*cptr = '\0';
		rid.isbranch = 0;
		rid.branch = "";
		rid.rev = prev;
		r = find_revision(rcs, &rid);
		if(!r)
		{
			/* Hm, disjoint branch... */
			fprintf(stderr, "Hm, don't know how to handle disjoint branches (%s)\n", prev);
			assert(r != NULL);
		}
		rcs->branches[n]->branchpoint = r;
		r->branches = xrealloc(r->branches, (r->nbranches+1)*sizeof(r->branches[0]));
		r->branches[r->nbranches] = rcs->branches[n];
		r->nbranches++;
	}

	make_layout(rcs);

#ifdef DEBUG
	dump_log(rcs);
#endif
	im = make_image(rcs);
	if(outfile)
	{
		if((fp = fopen(outfile, "w")) == NULL)
		{
			perror(outfile);
			return 1;
		}
	}
	else
		fp = stdout;

	switch(conf.image_type)
	{
#ifdef HAVE_IMAGE_GIF
	default:
	case IMAGE_GIF:
		gdImageGif(im, fp);
		break;
#endif
#ifdef HAVE_IMAGE_PNG
# ifndef HAVE_IMAGE_GIF
	default:
# endif
	case IMAGE_PNG:
		gdImagePng(im, fp);
		break;
#endif
#ifdef HAVE_IMAGE_JPEG
# if !defined(HAVE_IMAGE_GIF) && !defined(HAVE_IMAGE_PNG)
	default:
# endif
	case IMAGE_JPEG:
		gdImageJpeg(im, fp, conf.image_quality);
		break;
#endif
	}

	if(outfile)
		fclose(fp);
	gdImageDestroy(im);
	return 0;
}

